from interactions import (
    ApplicationCommandType,
    Guild,
    Option,
    InteractionException,
    ApplicationCommand,
)
from typing import Coroutine, Optional, Union, List, Dict, Any, Callable
from interactions.decor import command


def message_menu(
    self,
    *,
    name: Optional[str] = None,
    scope: Optional[Union[int, Guild, List[int], List[Guild]]] = None,
    default_permission: Optional[bool] = None,
) -> Callable[..., Any]:
    """
    A decorator for registering a message context menu to the Discord API,
    as well as being able to listen for ``INTERACTION_CREATE`` dispatched
    gateway events.
    The structure of a user context menu:
    .. code-block:: python
        @user_menu(name="Context menu name", description="this is a message context menu.")
        async def context_menu_name(ctx):
            ...
    The ``scope`` kwarg field may also be used to designate the command in question
    applicable to a guild or set of guilds.
    :param name: The name of the application command. This *is* required but kept optional to follow kwarg rules.
    :type name: Optional[str]
    :param scope?: The "scope"/applicable guilds the application command applies to.
    :type scope: Optional[Union[int, Guild, List[int], List[Guild]]]
    :param default_permission?: The default permission of accessibility for the application command. Defaults to ``True``.
    :type default_permission: Optional[bool]
    :return: A callable response.
    :rtype: Callable[..., Any]
    """

    def decorator(coro: Coroutine) -> Callable[..., Any]:
        if not name:
            raise InteractionException(11, message="Your command must have a name.")

        if not len(coro.__code__.co_varnames):
            raise InteractionException(
                11,
                message="Your command needs at least one argument to return context.",
            )

        commands: List[ApplicationCommand] = command(
            type=ApplicationCommandType.MESSAGE,
            name=name,
            description=None,
            scope=scope,
            options=None,
            default_permission=default_permission,
        )

        if self.automate_sync:
            [
                self.loop.run_until_complete(self.synchronize(command))
                for command in commands
            ]

        return self.event(coro, name=f"command_{name}")

    return decorator


def user_menu(
    self,
    *,
    name: Optional[str] = None,
    scope: Optional[Union[int, Guild, List[int], List[Guild]]] = None,
    default_permission: Optional[bool] = None,
) -> Callable[..., Any]:
    """
    A decorator for registering a user context menu to the Discord API,
    as well as being able to listen for ``INTERACTION_CREATE`` dispatched
    gateway events.
    The structure of a user context menu:
    .. code-block:: python
        @user_menu(name="Context menu name", description="this is a user context menu.")
        async def context_menu_name(ctx):
            ...
    The ``scope`` kwarg field may also be used to designate the command in question
    applicable to a guild or set of guilds.
    :param name: The name of the application command. This *is* required but kept optional to follow kwarg rules.
    :type name: Optional[str]
    :param scope?: The "scope"/applicable guilds the application command applies to.
    :type scope: Optional[Union[int, Guild, List[int], List[Guild]]]
    :param default_permission?: The default permission of accessibility for the application command. Defaults to ``True``.
    :type default_permission: Optional[bool]
    :return: A callable response.
    :rtype: Callable[..., Any]
    """

    def decorator(coro: Coroutine) -> Callable[..., Any]:
        if not name:
            raise InteractionException(11, message="Your command must have a name.")

        if not len(coro.__code__.co_varnames):
            raise InteractionException(
                11,
                message="Your command needs at least one argument to return context.",
            )

        commands: List[ApplicationCommand] = command(
            type=ApplicationCommandType.USER,
            name=name,
            description=None,
            scope=scope,
            options=None,
            default_permission=default_permission,
        )

        if self.automate_sync:
            [
                self.loop.run_until_complete(self.synchronize(command))
                for command in commands
            ]

        return self.event(coro, name=f"command_{name}")

    return decorator



from interactions import Button as B
from interactions import SelectMenu as SM
from interactions import ActionRow as AR
from interactions import ButtonStyle, Emoji, SelectOption
from typing import Union, List, Optional


def ActionRow(*args: Union[B, SM]) -> AR:
    """
    A helper function that passes arguments to `ActionRow`

    :param *args: The components to add to the ActionRow
    :type *args: Union[Button, SelectMenu]
    :return: The ActionRow
    :rtype: ActionRow
    """
    return AR(components=list(args))


def Button(
    style: Union[ButtonStyle, int],
    label: str,
    *,
    custom_id: Optional[str] = None,
    url: Optional[str] = None,
    emoji: Optional[Emoji] = None,
    disabled: Optional[bool] = False,
) -> B:
    """
    A helper function that passes arguments to `Button`

    :param style: The style of the button
    :type style: Union[ButtonStyle, int]
    :param label: The label of the button
    :type label: str
    :param custom_id: The custom id of the Button
    :type custom_id: Optional[str]
    :param url: The url of the Button
    :type url: Optional[str]
    :param emoji: The emoji of the Button
    :type emoji: Optional[Emoji]
    :param disabled: Whether the Button is disabled
    :type disabled: Optional[bool]
    :return: The Button
    :rtype: Button
    """
    if custom_id and url:
        raise ValueError("`custom_id` and `url` cannot be used together!")

    if not (custom_id or url):
        raise ValueError("`custom_id` or `url` must be specified!")

    if style == ButtonStyle.LINK and not url:
        raise ValueError("`url` must be specified if `style` is `ButtonStyle.LINK`!")
    elif url and style != ButtonStyle.LINK:
        raise ValueError(
            "`url` can only be specified if `style` is `ButtonStyle.LINK`!"
        )

    if style != ButtonStyle.LINK and not custom_id:
        raise ValueError(
            "`custom_id` must be specified if `style` is not `ButtonStyle.LINK`!"
        )
    elif custom_id and style == ButtonStyle.LINK:
        raise ValueError(
            "`custom_id` can only be specified if `style` is not `ButtonStyle.LINK`!"
        )

    return B(
        style=style,
        label=label,
        custom_id=custom_id,
        url=url,
        emoji=emoji,
        disabled=disabled,
    )


def SelectMenu(
    options: List[SelectOption],
    custom_id: str,
    *,
    placeholder: Optional[str] = None,
    min_values: Optional[int] = None,
    max_values: Optional[int] = None,
    disabled: Optional[bool] = False,
) -> SM:
    """
    A helper function that passes arguments to `SelectMenu`

    :param options: The options to display
    :type options: List[SelectOption]
    :param custom_id: The custom id of the SelectMenu
    :type custom_id: str
    :param placeholder: The placeholder of the SelectMenu
    :type placeholder: Optional[str]
    :param min_values: The minimum number of values that can be selected
    :type min_values: Optional[int]
    :param max_values: The maximum number of values that can be selected
    :type max_values: Optional[int]
    :param disabled: Whether the SelectMenu is disabled
    :type disabled: Optional[bool]
    :return: The SelectMenu
    :rtype: SelectMenu
    """
    return SM(
        options=options,
        custom_id=custom_id,
        placeholder=placeholder,
        min_values=min_values,
        max_values=max_values,
        disabled=disabled,
    )


def spread_to_rows(*components: Union[AR, B, SM], max_in_row: int = 5) -> List[AR]:
    """
    A helper function that spreads your components into `ActionRow`s of a set size

    :param *components: The components to spread, use `None` to explicit start a new row
    :type *components: Union[AR, Button, SelectMenu]
    :param max_in_row: The maximum number of components in each row
    :type max_in_row: int
    :return: The components spread to rows
    :rtype: List[ActionRow]
    :raises: ValueError: Too many or few components or rows
    """
    # todo: incorrect format errors
    if not components or len(components) > 25:
        raise ValueError("Number of components should be between 1 and 25.")
    if not 1 <= max_in_row <= 5:
        raise ValueError("max_in_row should be between 1 and 5.")

    rows = []
    action_row = []
    for component in list(components):
        if component is not None and isinstance(component, B):
            action_row.append(component)

            if len(action_row) == max_in_row:
                rows.append(ActionRow(*action_row))
                action_row = []

            continue

        if action_row:
            rows.append(ActionRow(*action_row))
            action_row = []

        if component is not None:
            if isinstance(component, AR):
                rows.append(component)
            elif isinstance(component, SM):
                rows.append(ActionRow(component))
    if action_row:
        rows.append(ActionRow(*action_row))

    if len(rows) > 5:
        raise ValueError("Number of rows exceeds 5.")

    return rows



import types
import interactions
from interactions.ext import wait_for

from .callback import component
from .context_menu import message_menu, user_menu
from .subcomand import base


class ExtendedWebSocket(interactions.api.gateway.WebSocket):
    def handle_dispatch(self, event: str, data: dict) -> None:
        super().handle_dispatch(event, data)

        if event == "INTERACTION_CREATE":
            if "type" not in data:
                return

            context: interactions.ComponentContext = self.contextualize(data)

            # startswith component callbacks
            if context.data.custom_id:
                for event in self.dispatch.events:
                    try:
                        startswith = self.dispatch.events[event][0].startswith
                    except AttributeError:
                        continue
                    if startswith and context.data.custom_id.startswith(
                        event.replace("component_startswith_", "")
                    ):
                        self.dispatch.dispatch(event, context)


interactions.api.gateway.WebSocket = ExtendedWebSocket


def _replace_values(old, new):
    """Change all values on new to the values on old. Useful if neither object has __dict__"""
    for item in dir(old):  # can't use __dict__, this should take everything
        value = getattr(old, item)

        if hasattr(value, "__call__") or isinstance(value, property):
            # Don't need to get callables or properties, that would un-overwrite things
            continue

        try:
            new.__setattr__(item, value)
        except AttributeError:
            pass


def setup(
    bot: interactions.Client,
    modify_component_callbacks: bool = True,
    add_context_menu: bool = True,
    add_subcommand: bool = True,
    add_method: bool = False,
    add_interaction_events: bool = False,
) -> None:
    """
    Apply hooks to a bot to add additional features

    This function is required, as importing alone won't extend the classes

    :param Client bot: The bot instance or class to apply hooks to
    :param bool modify_component_callbacks: Whether to modify the component callbacks
    :param bool add_method: If ``wait_for`` should be attached to the bot
    :param bool add_interaction_events: Whether to add ``on_message_component``, ``on_application_command``, and other interaction event
    """

    if not isinstance(bot, interactions.Client):
        raise TypeError(f"{bot.__class__.__name__} is not interactions.Client!")

    if modify_component_callbacks:
        bot.component = types.MethodType(component, bot)

        old_websocket = bot.websocket
        new_websocket = ExtendedWebSocket(
            old_websocket.intents, old_websocket.session_id, old_websocket.sequence
        )

        _replace_values(old_websocket, new_websocket)

        bot.websocket = new_websocket

    if add_context_menu:
        bot.message_menu = types.MethodType(message_menu, bot)
        bot.user_menu = types.MethodType(user_menu, bot)

    if add_subcommand:
        bot.base = types.MethodType(base, bot)

    if add_method or add_interaction_events:
        wait_for.setup(
            bot, add_method=add_method, add_interaction_events=add_interaction_events
        )



from typing import Union, Callable, Optional, Any, Coroutine
import interactions


def component(
    bot: interactions.Client,
    component: Union[str, interactions.Button, interactions.SelectMenu],
    startswith: Optional[bool] = False,
) -> Callable[..., Any]:
    """
    A decorator for listening to ``INTERACTION_CREATE`` dispatched gateway
    events involving components.
    The structure for a component callback:
    .. code-block:: python
        # Method 1
        @component(interactions.Button(
            style=interactions.ButtonStyle.PRIMARY,
            label="click me!",
            custom_id="click_me_button",
        ))
        async def button_response(ctx):
            ...
        # Method 2
        @component("custom_id")
        async def button_response(ctx):
            ...
    The context of the component callback decorator inherits the same
    as of the command decorator.
    :param component: The component you wish to callback for.
    :type component: Union[str, Button, SelectMenu]
    :param startswith: Whether the component should be matched by the start of the custom_id.
    :type startswith: bool
    :return: A callable response.
    :rtype: Callable[..., Any]
    """

    def decorator(coro: Coroutine) -> Any:
        payload: str = (
            interactions.Component(**component._json).custom_id
            if isinstance(component, (interactions.Button, interactions.SelectMenu))
            else component
        )
        if not startswith:
            coro.startswith = False
            return bot.event(coro, name=f"component_{payload}")
        coro.startswith = True
        return bot.event(coro, name=f"component_startswith_{payload}")

    return decorator



from typing import Union, Coroutine, Optional, List, Dict, Any, Callable
from interactions import (
    ApplicationCommandType,
    Client,
    Guild,
    Option,
    InteractionException,
    ApplicationCommand,
    OptionType,
)
from interactions.decor import command


class Subcommand:
    def __init__(
        self,
        name: str,
        description: str,
        coro: Coroutine,
        options: List[Option] = None,
    ):
        self.name: str = name
        self.description: str = description
        self.coro: Coroutine = coro
        self.options: List[Option] = options
        self._options: Option = Option(
            type=OptionType.SUB_COMMAND,
            name=name,
            description=description,
            options=options,
        )


class Group:
    def __init__(
        self, group: str, description: str, subcommand: Optional[Subcommand] = None
    ):
        self.group: str = group
        self.description: str = description
        self.subcommands: List[Subcommand] = [] if not subcommand else [subcommand]

    @property
    def _options(self) -> Option:
        return Option(
            type=OptionType.SUB_COMMAND_GROUP,
            name=self.group,
            description=self.description,
            options=[subcommand._options for subcommand in self.subcommands]
            if self.subcommands
            else None,
        )


class SubcommandSetup:
    def __init__(
        self,
        client: Client,
        base: str,
        description: Optional[str] = "No description",
        scope: Optional[Union[int, Guild, List[int], List[Guild]]] = None,
        default_permission: Optional[bool] = None,
    ):
        self.client: Client = client
        self.base: str = base
        self.description: str = description
        self.scope: Union[int, Guild, List[int], List[Guild]] = scope
        self.default_permission: bool = default_permission

        self.groups: Dict[str, Group] = {}
        self.subcommands: Dict[str, Subcommand] = {}

    def subcommand(
        self,
        *,
        group: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        options: Optional[List[Option]] = None,
    ):
        def decorator(coro: Coroutine):
            if not name:
                raise InteractionException(
                    11, message="Your command must have a base and name."
                )

            if not description:
                raise InteractionException(
                    11, message="Chat-input commands must have a description."
                )

            if not len(coro.__code__.co_varnames):
                raise InteractionException(
                    11,
                    message="Your command needs at least one argument to return context.",
                )
            if options and (len(coro.__code__.co_varnames) + 1) < len(options):
                raise InteractionException(
                    11,
                    message="You must have the same amount of arguments as the options of the command.",
                )

            if group:
                if group not in self.groups:
                    self.groups[group] = Group(
                        group,
                        description,
                        subcommand=Subcommand(name, description, coro, options),
                    )
                else:
                    subcommands = self.groups[group].subcommands
                    subcommands.append(Subcommand(name, description, coro, options))
            else:
                self.subcommands[name] = Subcommand(name, description, coro, options)

            return coro

        return decorator

    def finish(self):
        options = (
            [group._options for group in self.groups.values()]
            + [subcommand._options for subcommand in self.subcommands.values()]
        ) or None
        commands: List[ApplicationCommand] = command(
            type=ApplicationCommandType.CHAT_INPUT,
            name=self.base,
            description=self.description,
            scope=self.scope,
            options=options,
        )

        if self.client.automate_sync:
            [
                self.client.loop.run_until_complete(self.client.synchronize(command))
                for command in commands
            ]

        async def inner(ctx, *args, sub_command_group=None, sub_command=None, **kwargs):
            if sub_command_group:
                group = self.groups[sub_command_group]
                for subcommand in group.subcommands:
                    if subcommand.name == sub_command:
                        break
            else:
                subcommand = self.subcommands[sub_command]

            original_coro = subcommand.coro
            return await original_coro(ctx, *args, **kwargs)

        return self.client.event(inner, name=f"command_{self.base}")


def base(
    self: Client,
    base: str,
    description: Optional[str] = "No description",
    scope: Optional[Union[int, Guild, List[int], List[Guild]]] = None,
    default_permission: Optional[bool] = None,
):
    return SubcommandSetup(self, base, description, scope, default_permission)



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="better-interactions",
    version="2.0.1",
    description="Better interactions for discord-py-interactions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Toricane/better-interactions",
    author="Toricane",
    author_email="prjwl028@gmail.com",
    license="MIT",
    packages=["interactions.ext.better_interactions"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "discord-py-interactions",
        "interactions-wait-for",
    ],
)
